@@ -33,8 +33,6 @@
 namespace ar
 {
 
-typedef std::auto_ptr<fastq_read_chunk> chunk_ptr;
-
 
 size_t read_fastq_reads(fastq_vec& dst, line_reader& reader, size_t offset,
                       const fastq_encoding& encoding)
@@ -129,7 +127,7 @@ chunk_vec read_single_fastq::process(analytical_chunk* chunk)
         return chunk_vec();
     }
 
-    chunk_ptr file_chunk(new fastq_read_chunk());
+    read_chunk_ptr file_chunk(new fastq_read_chunk());
 
     const size_t n_read = read_fastq_reads(file_chunk->reads_1, m_io_input,
                                            m_line_offset, *m_encoding);
@@ -145,7 +143,7 @@ chunk_vec read_single_fastq::process(analytical_chunk* chunk)
     m_line_offset += n_read;
 
     chunk_vec chunks;
-    chunks.push_back(chunk_pair(m_next_step, file_chunk.release()));
+    chunks.push_back(chunk_pair(m_next_step, std::move(file_chunk)));
 
     return chunks;
 }
@@ -184,7 +182,7 @@ chunk_vec read_paired_fastq::process(analytical_chunk* chunk)
         return chunk_vec();
     }
 
-    chunk_ptr file_chunk(new fastq_read_chunk());
+    read_chunk_ptr file_chunk(new fastq_read_chunk());
 
     const size_t n_read_1 = read_fastq_reads(file_chunk->reads_1, m_io_input_1,
                                              m_line_offset, *m_encoding);
@@ -211,7 +209,7 @@ chunk_vec read_paired_fastq::process(analytical_chunk* chunk)
     m_line_offset += n_read_1;
 
     chunk_vec chunks;
-    chunks.push_back(chunk_pair(m_next_step, file_chunk.release()));
+    chunks.push_back(chunk_pair(m_next_step, std::move(file_chunk)));
 
     return chunks;
 }
@@ -249,7 +247,7 @@ chunk_vec read_interleaved_fastq::process(analytical_chunk* chunk)
         return chunk_vec();
     }
 
-    chunk_ptr file_chunk(new fastq_read_chunk());
+    read_chunk_ptr file_chunk(new fastq_read_chunk());
 
     file_chunk->reads_1.reserve(FASTQ_CHUNK_SIZE);
     file_chunk->reads_2.reserve(FASTQ_CHUNK_SIZE);
@@ -304,7 +302,7 @@ chunk_vec read_interleaved_fastq::process(analytical_chunk* chunk)
     m_line_offset += (n_read_1 + n_read_2) * 4;
 
     chunk_vec chunks;
-    chunks.push_back(chunk_pair(m_next_step, file_chunk.release()));
+    chunks.push_back(chunk_pair(m_next_step, std::move(file_chunk)));
 
     return chunks;
 }
@@ -345,10 +343,10 @@ std::pair<size_t, unsigned char*> build_input_buffer(const string_vec& lines)
 #ifdef AR_BZIP2_SUPPORT
 
 ///////////////////////////////////////////////////////////////////////////////
-// Implementations for 'gzip_paired_fastq'
+// Implementations for 'gzip_fastq'
 
 
-bzip2_paired_fastq::bzip2_paired_fastq(const userconfig& config, size_t next_step)
+bzip2_fastq::bzip2_fastq(const userconfig& config, size_t next_step)
   : analytical_step(analytical_step::ordered, false)
   , m_buffered_reads(0)
   , m_next_step(next_step)
@@ -369,24 +367,24 @@ bzip2_paired_fastq::bzip2_paired_fastq(const userconfig& config, size_t next_ste
             break;
 
         case BZ_MEM_ERROR:
-            throw thread_error("bzip2_paired_fastq: not enough memory");
+            throw thread_error("bzip2_fastq: not enough memory");
 
         case BZ_CONFIG_ERROR:
-            throw thread_error("bzip2_paired_fastq: miscompiled bzip2 library");
+            throw thread_error("bzip2_fastq: miscompiled bzip2 library");
 
         case BZ_PARAM_ERROR:
-            throw thread_error("bzip2_paired_fastq: invalid parameters");
+            throw thread_error("bzip2_fastq: invalid parameters");
 
         default:
-            throw thread_error("bzip2_paired_fastq: unknown error");
+            throw thread_error("bzip2_fastq: unknown error");
     }
 }
 
 
-void bzip2_paired_fastq::finalize()
+void bzip2_fastq::finalize()
 {
     if (!m_eof) {
-        throw thread_error("bzip2_paired_fastq::finalize: terminated before EOF");
+        throw thread_error("bzip2_fastq::finalize: terminated before EOF");
     }
 
     const int errorcode = BZ2_bzCompressEnd(&m_stream);
@@ -394,31 +392,27 @@ void bzip2_paired_fastq::finalize()
         print_locker lock;
         switch (errorcode) {
             case BZ_PARAM_ERROR:
-                throw thread_error("bzip2_paired_fastq::finalize: parameter error");
+                throw thread_error("bzip2_fastq::finalize: parameter error");
 
             default:
-                throw thread_error("Unknown error in bzip2_paired_fastq::finalize");
+                throw thread_error("Unknown error in bzip2_fastq::finalize");
         }
     }
 }
 
 
-chunk_vec bzip2_paired_fastq::process(analytical_chunk* chunk)
+chunk_vec bzip2_fastq::process(analytical_chunk* chunk)
 {
-    std::auto_ptr<fastq_output_chunk> file_chunk(dynamic_cast<fastq_output_chunk*>(chunk));
+    output_chunk_ptr file_chunk(dynamic_cast<fastq_output_chunk*>(chunk));
     buffer_vec& buffers = file_chunk->buffers;
 
     if (m_eof) {
-        throw thread_error("bzip2_paired_fastq::process: received data after EOF");
+        throw thread_error("bzip2_fastq::process: received data after EOF");
     }
 
     m_eof = file_chunk->eof;
     if (file_chunk->reads.empty() && !m_eof) {
-        // The empty chunk must still be forwarded, to ensure that tracking of
-        // ordered chunks does not break.
-        chunk_vec chunks;
-        chunks.push_back(chunk_pair(m_next_step, file_chunk.release()));
-        return chunks;
+        return chunk_vec();
     }
 
     std::pair<size_t, unsigned char*> input_buffer;
@@ -447,16 +441,16 @@ chunk_vec bzip2_paired_fastq::process(analytical_chunk* chunk)
                         break;
 
                     case BZ_FLUSH_OK:
-                        throw thread_error("bzip2_paired_fastq::process: BZ_FLUSH_OK");
+                        throw thread_error("bzip2_fastq::process: BZ_FLUSH_OK");
 
                     case BZ_PARAM_ERROR:
-                        throw thread_error("bzip2_paired_fastq::process: BZ_PARAM_ERROR");
+                        throw thread_error("bzip2_fastq::process: BZ_PARAM_ERROR");
 
                     case BZ_SEQUENCE_ERROR:
-                        throw thread_error("bzip2_paired_fastq::process: sequence error");
+                        throw thread_error("bzip2_fastq::process: sequence error");
 
                     default:
-                        throw thread_error("bzip2_paired_fastq::process: unknown error");
+                        throw thread_error("bzip2_fastq::process: unknown error");
                 }
 
                 output_buffer.first = FASTQ_COMPRESSED_CHUNK - m_stream.avail_out;
@@ -480,7 +474,7 @@ chunk_vec bzip2_paired_fastq::process(analytical_chunk* chunk)
     chunk_vec chunks;
     if (!file_chunk->buffers.empty() || m_eof) {
         file_chunk->count += m_buffered_reads;
-        chunks.push_back(chunk_pair(m_next_step, file_chunk.release()));
+        chunks.push_back(chunk_pair(m_next_step, std::move(file_chunk)));
         m_buffered_reads = 0;
     } else {
         m_buffered_reads += file_chunk->count;
@@ -495,9 +489,9 @@ chunk_vec bzip2_paired_fastq::process(analytical_chunk* chunk)
 #ifdef AR_GZIP_SUPPORT
 
 ///////////////////////////////////////////////////////////////////////////////
-// Implementations for 'gzip_paired_fastq'
+// Implementations for 'gzip_fastq'
 
-gzip_paired_fastq::gzip_paired_fastq(const userconfig& config, size_t next_step)
+gzip_fastq::gzip_fastq(const userconfig& config, size_t next_step)
   : analytical_step(analytical_step::ordered, false)
   , m_buffered_reads(0)
   , m_next_step(next_step)
@@ -520,24 +514,24 @@ gzip_paired_fastq::gzip_paired_fastq(const userconfig& config, size_t next_step)
             break;
 
         case Z_MEM_ERROR:
-            throw thread_error("gzip_paired_fastq: not enough memory");
+            throw thread_error("gzip_fastq: not enough memory");
 
         case Z_STREAM_ERROR:
-            throw thread_error("gzip_paired_fastq: invalid parameters");
+            throw thread_error("gzip_fastq: invalid parameters");
 
         case Z_VERSION_ERROR:
-            throw thread_error("gzip_paired_fastq: incompatible zlib version");
+            throw thread_error("gzip_fastq: incompatible zlib version");
 
         default:
-            throw thread_error("gzip_paired_fastq: unknown error");
+            throw thread_error("gzip_fastq: unknown error");
     }
 }
 
 
-void gzip_paired_fastq::finalize()
+void gzip_fastq::finalize()
 {
     if (!m_eof) {
-        throw thread_error("gzip_paired_fastq::finalize: terminated before EOF");
+        throw thread_error("gzip_fastq::finalize: terminated before EOF");
     }
 
     const int errorcode = deflateEnd(&m_stream);
@@ -545,34 +539,30 @@ void gzip_paired_fastq::finalize()
         print_locker lock;
         switch (errorcode) {
             case Z_STREAM_ERROR:
-                throw thread_error("gzip_paired_fastq::finalize: stream error");
+                throw thread_error("gzip_fastq::finalize: stream error");
 
             case Z_DATA_ERROR:
-                throw thread_error("gzip_paired_fastq::finalize: data error");
+                throw thread_error("gzip_fastq::finalize: data error");
 
             default:
-                throw thread_error("Unknown error in gzip_paired_fastq::finalize");
+                throw thread_error("Unknown error in gzip_fastq::finalize");
         }
     }
 }
 
 
-chunk_vec gzip_paired_fastq::process(analytical_chunk* chunk)
+chunk_vec gzip_fastq::process(analytical_chunk* chunk)
 {
-    std::auto_ptr<fastq_output_chunk> file_chunk(dynamic_cast<fastq_output_chunk*>(chunk));
+    output_chunk_ptr file_chunk(dynamic_cast<fastq_output_chunk*>(chunk));
     buffer_vec& buffers = file_chunk->buffers;
 
     if (m_eof) {
-        throw thread_error("bzip2_paired_fastq::process: received data after EOF");
+        throw thread_error("bzip2_fastq::process: received data after EOF");
     }
 
     m_eof = file_chunk->eof;
     if (file_chunk->reads.empty() && !m_eof) {
-        // The empty chunk must still be forwarded, to ensure that tracking of
-        // ordered chunks does not break.
-        chunk_vec chunks;
-        chunks.push_back(chunk_pair(m_next_step, file_chunk.release()));
-        return chunks;
+        return chunk_vec();
     }
 
     std::pair<size_t, unsigned char*> input_buffer;
@@ -600,13 +590,13 @@ chunk_vec gzip_paired_fastq::process(analytical_chunk* chunk)
                         break;
 
                     case Z_BUF_ERROR:
-                        throw thread_error("gzip_paired_fastq::process: buf error");
+                        throw thread_error("gzip_fastq::process: buf error");
 
                     case Z_STREAM_ERROR:
-                        throw thread_error("gzip_paired_fastq::process: stream error");
+                        throw thread_error("gzip_fastq::process: stream error");
 
                     default:
-                        throw thread_error("gzip_paired_fastq::process: unknown error");
+                        throw thread_error("gzip_fastq::process: unknown error");
                 }
 
                 output_buffer.first = FASTQ_COMPRESSED_CHUNK - m_stream.avail_out;
@@ -630,7 +620,7 @@ chunk_vec gzip_paired_fastq::process(analytical_chunk* chunk)
     chunk_vec chunks;
     if (!file_chunk->buffers.empty() || m_eof) {
         file_chunk->count += m_buffered_reads;
-        chunks.push_back(chunk_pair(m_next_step, file_chunk.release()));
+        chunks.push_back(chunk_pair(m_next_step, std::move(file_chunk)));
         m_buffered_reads = 0;
     } else {
         m_buffered_reads += file_chunk->count;
@@ -646,7 +636,7 @@ chunk_vec gzip_paired_fastq::process(analytical_chunk* chunk)
 // Implementations for 'write_fastq'
 
 //! Mutex used to control access to s_timer and s_finalized;
-static mutex s_timer_lock;
+static std::mutex s_timer_lock;
 //! Timer used to track trimming progress; accessed by all instances
 static timer s_timer = timer("reads");
 //! Indicates if 'timer::finalize' has been called.
@@ -655,7 +645,7 @@ static bool s_finalized = false;
 
 write_fastq::write_fastq(const std::string& filename)
   : analytical_step(analytical_step::ordered, true)
-  , m_output(filename.c_str(), std::ofstream::out)
+  , m_output(filename.c_str(), std::ofstream::out | std::ofstream::binary)
   , m_eof(false)
 {
     if (!m_output.is_open()) {
@@ -669,7 +659,7 @@ write_fastq::write_fastq(const std::string& filename)
 
 chunk_vec write_fastq::process(analytical_chunk* chunk)
 {
-    std::auto_ptr<fastq_output_chunk> file_chunk(dynamic_cast<fastq_output_chunk*>(chunk));
+    output_chunk_ptr file_chunk(dynamic_cast<fastq_output_chunk*>(chunk));
     const string_vec& lines = file_chunk->reads;
 
     if (m_eof) {
@@ -694,7 +684,7 @@ chunk_vec write_fastq::process(analytical_chunk* chunk)
         m_output.flush();
     }
 
-    mutex_locker lock(s_timer_lock);
+    std::lock_guard<std::mutex> lock(s_timer_lock);
     s_timer.increment(file_chunk->count);
 
     return chunk_vec();
@@ -703,7 +693,7 @@ chunk_vec write_fastq::process(analytical_chunk* chunk)
 
 void write_fastq::finalize()
 {
-    mutex_locker lock(s_timer_lock);
+    std::lock_guard<std::mutex> lock(s_timer_lock);
     if (!s_finalized) {
         s_timer.finalize();
         s_finalized = true;
