@@ -30,14 +30,17 @@
 #include <sys/time.h>
 #include <limits>
 
-#include "userconfig.h"
-#include "fastq.h"
 #include "alignment.h"
+#include "debug.h"
+#include "fastq.h"
 #include "strutils.h"
+#include "userconfig.h"
+
 
 namespace ar
 {
 
+
 size_t get_seed()
 {
     struct timeval timestamp;
@@ -47,11 +50,11 @@ size_t get_seed()
 }
 
 
-std::auto_ptr<fastq_encoding> select_encoding(const std::string& name,
-                                              const std::string& value,
-                                              size_t quality_max = MAX_PHRED_SCORE_DEFAULT)
+fastq_encoding_ptr select_encoding(const std::string& name,
+                                   const std::string& value,
+                                   size_t quality_max = MAX_PHRED_SCORE_DEFAULT)
 {
-    std::auto_ptr<fastq_encoding> ptr;
+    fastq_encoding_ptr ptr;
 
     const std::string uppercase_value = toupper(value);
     if (uppercase_value == "33") {
@@ -73,13 +76,14 @@ std::auto_ptr<fastq_encoding> select_encoding(const std::string& name,
 userconfig::userconfig(const std::string& name,
                        const std::string& version,
                        const std::string& help)
-    : argparser(name, version, help)
+    : run_type(ar_trim_adapters)
     , basename("your_output")
     , input_file_1()
     , input_file_2()
     , paired_ended_mode(false)
     , interleaved_input(false)
     , interleaved_output(false)
+    , combined_output(false)
     , mate_separator(MATE_SEPARATOR)
     , min_genomic_length(15)
     , max_genomic_length(std::numeric_limits<unsigned>::max())
@@ -95,7 +99,6 @@ userconfig::userconfig(const std::string& name,
     , collapse(false)
     , shift(2)
     , seed(get_seed())
-    , identify_adapters(false)
     , max_threads(1)
     , gzip(false)
     , gzip_level(6)
@@ -105,6 +108,7 @@ userconfig::userconfig(const std::string& name,
     , barcode_mm_r1(0)
     , barcode_mm_r2(0)
     , adapters()
+    , argparser(name, version, help)
     , adapter_1("AGATCGGAAGAGCACACGTCTGAACTCCAGTCACNNNNNNATCTCGTATGCCGTCTTCTGCTTG")
     , adapter_2("AGATCGGAAGAGCGTCGTGTAGGGAAAGAGTGTAGATCTCGGTGGTCGCCGTATCATT")
     , adapter_list()
@@ -114,6 +118,8 @@ userconfig::userconfig(const std::string& name,
     , quality_max(MAX_PHRED_SCORE_DEFAULT)
     , mate_separator_str(1, MATE_SEPARATOR)
     , interleaved(false)
+    , identify_adapters(false)
+    , demultiplex_sequences(false)
 {
     argparser["--file1"] =
         new argparse::any(&input_file_1, "FILE",
@@ -161,7 +167,14 @@ userconfig::userconfig(const std::string& name,
             "containing mate 1 and mate 2 reads, one pair after the other. "
             "This option is implied by the --interleaved option [current: "
             "%default].");
-
+    argparser["--combined-output"] =
+        new argparse::flag(&combined_output,
+            "If set, all reads are written to the same file(s), specified by "
+            "--output1 and --output2 (--output1 only if --interleaved-output "
+            "is not set). Each read is futher marked by either a \"PASSED\" "
+            "or a \"FAILED\" flag, and any read that has been FAILED "
+            "(including the mate for collapsed reads) are replaced with a "
+            "single 'N' with Phred score 0 [current: %default].");
 
     argparser.add_header("OUTPUT FILES:");
     argparser["--basename"] =
@@ -205,7 +218,6 @@ userconfig::userconfig(const std::string& name,
             "Contains reads discarded due to the --minlength, --maxlength or "
             "--maxns options [default: BASENAME.discarded]");
 
-
 #if defined(AR_GZIP_SUPPORT) || defined(AR_BZIP2_SUPPORT)
    argparser.add_header("OUTPUT COMPRESSION:");
 #endif
@@ -322,7 +334,6 @@ userconfig::userconfig(const std::string& name,
             "both single-end and paired-end trimming, if double-indexed "
             "multiplexing was used, in order to ensure that the demultiplexed "
             "reads can be trimmed correctly [current: %default].");
-
     argparser["--barcode-mm"] =
         new argparse::knob(&barcode_mm, "N",
             "Maximum number of mismatches allowed when counting mismatches in "
@@ -337,6 +348,11 @@ userconfig::userconfig(const std::string& name,
             "Maximum number of mismatches allowed for the mate 2 barcode; "
             "if not set, this value is equal to the '--barcode-mm' value; "
             "cannot be higher than the '--barcode-mm value'.");
+    argparser["--demultiplex-only"] =
+        new argparse::flag(&demultiplex_sequences,
+            "Only carry out demultiplexing using the list of barcodes "
+            "supplied with --barcode-list; do not attempt to trim adapters "
+            "to carry out other processing.");
 
     argparser.add_header("MISC:");
     argparser["--identify-adapters"] =
@@ -349,12 +365,9 @@ userconfig::userconfig(const std::string& name,
             "Phred scores when collapsing. Note that runs are not "
             "deterministic if more than one thread is used. If not specified, "
             "a seed is generated using the current time.");
-
-#ifdef AR_PTHREAD_SUPPORT
     argparser["--threads"] =
         new argparse::knob(&max_threads, "THREADS",
             "Maximum number of threads [current: %default]");
-#endif
 }
 
 
@@ -404,8 +417,28 @@ argparse::parse_result userconfig::parse_args(int argc, char *argv[])
         // sequences. However, arguments are still checked above.
         quality_input_fmt.reset(new fastq_encoding(PHRED_OFFSET_33, MAX_PHRED_SCORE));
         quality_output_fmt.reset(new fastq_encoding(PHRED_OFFSET_33, MAX_PHRED_SCORE));
+        run_type = ar_identify_adapters;
     }
 
+    if (demultiplex_sequences) {
+        if (identify_adapters) {
+            std::cerr << "Error: Cannot use --identify-adapters and "
+                      << "--demultiplex-only at the same time!"
+                      << std::endl;
+
+            return argparse::pr_error;
+        } else if (!argparser.is_set("--barcode-list")) {
+            std::cerr << "Error: Cannot use --demultiplex-only without specifying "
+                      << "a list of barcodes using --barcode-list!"
+                      << std::endl;
+
+            return argparse::pr_error;
+        }
+
+        run_type = ar_demultiplex_sequences;
+    }
+
+
     if (low_quality_score > static_cast<unsigned>(MAX_PHRED_SCORE)) {
         std::cerr << "Error: Invalid value for --minquality: "
                   << low_quality_score << "\n"
@@ -438,7 +471,8 @@ argparse::parse_result userconfig::parse_args(int argc, char *argv[])
 
     if (interleaved_input) {
         if (file_2_set) {
-            std::cerr << "Error: The option --interleaved cannot be used "
+            std::cerr << "Error: The options --interleaved and "
+                      << "--interleaved-input cannot be used "
                       << "together with the --file2 option; only --file1 must "
                       << "be specified!"
                       << std::endl;
@@ -498,27 +532,29 @@ argparse::parse_result userconfig::parse_args(int argc, char *argv[])
     if (!max_threads) {
         std::cerr << "Error: --threads must be at least 1!" << std::endl;
         return argparse::pr_error;
+    } else if (max_threads > 1 && argparser.is_set("--seed")) {
+        std::cerr << "Warning: The option --seed should not be used when "
+                  << "using multiple threads; multi-threaded behavior is not "
+                  << "deterministic even with a fixed seed!"
+                  << std::endl;
     }
 
     return argparse::pr_ok;
 }
 
 
-std::auto_ptr<statistics> userconfig::create_stats() const
+statistics_ptr userconfig::create_stats() const
 {
-    std::auto_ptr<statistics> stats(new statistics());
+    statistics_ptr stats(new statistics());
     stats->number_of_reads_with_adapter.resize(adapters.adapter_count());
     return stats;
 }
 
 
-userconfig::alignment_type userconfig::evaluate_alignment(const alignment_info& alignment) const
+bool userconfig::is_good_alignment(const alignment_info& alignment) const
 {
-    if (!alignment.length) {
-        return not_aligned;
-    } else if (alignment.score <= 0) {
-        // Very poor alignment, will not be considered
-        return poor_alignment;
+    if (!alignment.length || alignment.score <= 0) {
+        return false;
     }
 
     // Only pairs of called bases are considered part of the alignment
@@ -526,7 +562,7 @@ userconfig::alignment_type userconfig::evaluate_alignment(const alignment_info&
     size_t mm_threshold = static_cast<size_t>(mismatch_threshold * n_aligned);
 
     if (n_aligned < min_adapter_overlap) {
-        return poor_alignment;
+        return false;
     }
 
     if (n_aligned < 6) {
@@ -537,10 +573,10 @@ userconfig::alignment_type userconfig::evaluate_alignment(const alignment_info&
     }
 
     if (alignment.n_mismatches > mm_threshold) {
-        return not_aligned;
+        return false;
     }
 
-    return valid_alignment;
+    return true;
 }
 
 
@@ -580,9 +616,19 @@ std::string userconfig::get_output_filename(const std::string& key,
     } else if (key == "demux_unknown") {
         filename += ".unidentified";
 
-        if (nth) {
-            filename.push_back('_');
-            filename.push_back('0' + nth);
+        AR_DEBUG_ASSERT(nth <= 9);
+        if (!interleaved_output) {
+            if (nth) {
+                filename.push_back('_');
+                filename.push_back('0' + nth);
+            }
+        } else {
+            filename += ".paired";
+        }
+
+        // Currently only when demultiplexing; for backwards compatibility
+        if (run_type == ar_demultiplex_sequences) {
+            filename += ".fastq";
         }
 
         if (gzip) {
@@ -610,23 +656,32 @@ std::string userconfig::get_output_filename(const std::string& key,
     } else if (paired_ended_mode) {
         if (key == "--output1") {
             if (interleaved_output) {
-                filename += ".paired.truncated";
+                filename += ".paired";
             } else {
-                filename += ".pair1.truncated";
+                filename += ".pair1";
             }
         } else if (key == "--output2") {
-            filename += ".pair2.truncated";
+            filename += ".pair2";
         } else if (key == "--singleton") {
-            filename += ".singleton.truncated";
+            filename += ".singleton";
         } else {
             throw std::invalid_argument("invalid read-type in userconfig::get_output_filename constructor: " + key);
         }
-    } else {
-       if (key == "--output1") {
+
+        if (run_type != ar_demultiplex_sequences) {
             filename += ".truncated";
-        } else {
-            throw std::invalid_argument("invalid read-type in userconfig::get_output_filename constructor: " + key);
         }
+    } else if (key == "--output1") {
+        if (run_type != ar_demultiplex_sequences) {
+            filename += ".truncated";
+        }
+    } else if (key != "demux_unknown") {
+        throw std::invalid_argument("invalid read-type in userconfig::get_output_filename constructor: " + key);
+    }
+
+    // Currently only when demultiplexing; for backwards compatibility
+    if (run_type == ar_demultiplex_sequences) {
+        filename += ".fastq";
     }
 
     if (gzip) {
@@ -742,7 +797,7 @@ bool userconfig::setup_adapter_sequences()
         } else if (adapters.adapter_count()) {
             std::cerr << "Read " << adapters.barcode_count()
                       << " barcodes / barcode pairs from '" << barcode_list
-                      << "'..." << std::endl;
+                      << "' ..." << std::endl;
         } else {
             std::cerr << "Error: No barcodes sequences found in table!" << std::endl;
             return false;
